package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.ui.SceneAwareController;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Testbed;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.connection.*;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import com.sun.javafx.webkit.WebConsoleListener;
import javafx.application.Platform;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.concurrent.Worker;
import javafx.fxml.FXML;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Label;
import javafx.scene.control.ListView;
import javafx.scene.control.TextField;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyCodeCombination;
import javafx.scene.layout.BorderPane;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebView;
import javafx.stage.Stage;
import javafx.stage.Window;
import netscape.javascript.JSObject;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.*;
import org.apache.http.entity.BasicHttpEntity;
import org.apache.http.impl.client.HttpClientBuilder;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import java.io.*;
import java.net.*;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;

public class JFedBrowser implements SceneAwareController {
    private static final Logger LOG = LoggerFactory.getLogger(JFedBrowser.class);

    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final SfaConnectionPool connectionPool;
    @Nonnull private final JFedTrustStore trustStore;
    @Nonnull private final JFedPreferences preferences;

    @FXML protected WebView reservationWebView;
    @FXML protected Label headerLabel;
    @FXML protected Parent centerBox;
    @FXML protected TextField urlField;
    @FXML protected ListView<String> messageList;
    @FXML protected Label firebugLabel;

    protected WebEngine webEngine;
    protected boolean initialized;

    protected String initialUrl = null;

    protected Scene scene;
    protected Stage stage;

    protected int jsBridgeBackoff = 1;
    protected boolean forceCloseMessages = false;

    @Override
    public void registerScene(Scene scene) {
        this.scene = scene;
        this.stage = (Stage) scene.getWindow();

        LOG.debug("Loaded accelerators");
        scene.getAccelerators().put(
                new KeyCodeCombination(KeyCode.F12),
                this::openFirebug
        );
    }

    public void setInitialUrl(@Nullable String initialUrl) {
        if (initialUrl == null) {
            this.initialUrl = "https://www.wall2.ilabt.iminds.be:12369/ilabt/1.0/userprojects.php";
        } else {
            this.initialUrl = initialUrl;
        }

        loadPageWhenReady();
    }

    public static class JavaScriptBridge {
        @Nonnull protected final JFedBrowser browser;
        public JavaScriptBridge(@Nonnull JFedBrowser browser) {
            this.browser = browser;
        }

        public void approve() {
            LOG.debug("JavaScriptBridge.approve");
            Platform.runLater(() -> browser.messageList.getItems().add("-> jfed.approve called"));
        }
        public void decline() {
            LOG.debug("JavaScriptBridge.decline");
            Platform.runLater(() -> browser.messageList.getItems().add("-> jfed.decline called"));
        }
        public void close() {
            LOG.debug("JavaScriptBridge.close");
            Platform.runLater(() -> browser.messageList.getItems().add("-> jfed.close called"));
//            Window stage = browser.scene.getWindow();
//            stage.hide();
        }
    }

    @Inject
    public JFedBrowser(@Nonnull GeniUserProvider geniUserProvider,
                       @Nonnull SfaConnectionPool connectionPool,
                       @Nonnull JFedTrustStore trustStore,
                       @Nonnull JFedPreferences preferences) {
        this.geniUserProvider = geniUserProvider;
        this.connectionPool = connectionPool;
        this.trustStore = trustStore;
        this.preferences = preferences;
    }

    @FXML
    protected void initialize() {
        webEngine = reservationWebView.getEngine();

        URLStreamFactoryCustomizer.useDedicatedProxyForWebkit(
                new MyHttpUrlStreamHandler(), "https");
        initialized = true;

        headerLabel.setText("JFed browser with SSL-client authentication for user "+
                geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName());

        messageList.setItems(FXCollections.observableArrayList());
        messageList.getItems().addListener((ListChangeListener<String>) c -> messageList.setVisible(!c.getList().isEmpty() && !forceCloseMessages));

        webEngine.getLoadWorker().stateProperty().addListener(
                (ObservableValue<? extends Worker.State> ov, Worker.State oldState,
                 Worker.State newState) -> {
                    if (newState == Worker.State.SUCCEEDED) {
                        Platform.runLater(() -> urlField.setText(webEngine.getLocation()));
                        jsBridgeBackoff = 1;
                        registerJsBridge();
                    }
                });

        WebConsoleListener.setDefaultListener((webView, message, lineNumber, sourceId) -> {
            LOG.debug("[Webengine message] "+message + "[at " + lineNumber + "]");
            Platform.runLater(() -> messageList.getItems().add(message + "[at " + lineNumber + "]"));

        });

        loadPageWhenReady();
    }

    public void closeMessages() {
        forceCloseMessages = true;
        messageList.setVisible(!messageList.getItems().isEmpty() && !forceCloseMessages);
    }

    protected boolean injectedFirebug = false;
    protected void openFirebug() {
        if (injectedFirebug) {
            LOG.warn("Not injecting firebug again.");
            return;
        }
        String firebugScript;
        try {
            firebugScript = IOUtils.resourceToString("firebug-lite.js", this.getClass());
        } catch (IOException e) {
            throw new RuntimeException("Error loading firebug-lite.js", e);
        }
        LOG.debug("Injecting firebug");
        webEngine.executeScript(firebugScript);
        injectedFirebug = true;

        if (firebugLabel != null)
            firebugLabel.setVisible(false);


        //alternative
//        webEngine.executeScript("if (!document.getElementById('FirebugLite')){" +
//                "E = document['createElement' + 'NS'] && document.documentElement.namespaceURI;" +
//                "E = E ? document['createElement' + 'NS'](E, 'script') : document['createElement']('script');" +
//                "E['setAttribute']('id', 'FirebugLite');" +
//                "E['setAttribute']('type', 'text/javascript');" +
//                "E['setAttribute']('src', 'https://flsmonitor.fed4fire.eu/firebug-lite.js');" +
//                "(document['getElementsByTagName']('head')[0] || document['getElementsByTagName']('body')[0]).appendChild(E);" +
//                "}");
    }

    public void onUrlAction() {
        loadPage(urlField.getText());
    }

    public void close() {
        webEngine.getLoadWorker().cancel();
    }

    protected void loadPageWhenReady() {
        if (!initialized)
            return;

        if (initialUrl == null)
            return;

        urlField.setText(initialUrl);
        onUrlAction();
    }

    protected void loadPage(@Nonnull String urlString) {
        webEngine.load(urlString);
        jsBridgeBackoff = 1;
        Platform.runLater(this::registerJsBridge);
    }

    public void back() {
        webEngine.executeScript("history.back()");
    }

    public void forward() {
        webEngine.executeScript("history.forward()");
    }

    public void reload() {
        webEngine.reload();
        registerJsBridge();
//        loadPage(urlField.getText());
    }

    protected final ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(1);
    protected Future nextRegisterJsBridge = null;
    protected synchronized void registerJsBridgeLater() {
        Future snapshotOfFuture = nextRegisterJsBridge;
        if (snapshotOfFuture != null) {
            //try to stop previous task. No problem if it doesn't stop.
            //any race conditions here are also not important.
            snapshotOfFuture.cancel(false);
        }
//        LOG.debug("Trying registerJsBridge() again later.");
        int multi = jsBridgeBackoff;
        //a simple exponential backoff like thing
        if (multi < 1) multi = 1;
        if (multi > 4) multi = 4;
        multi = 1 << (multi-1);
//        LOG.debug("jsBridgeBackoff="+jsBridgeBackoff+" multi="+multi+" delay="+(250L * multi));
        nextRegisterJsBridge = scheduler.schedule(() -> Platform.runLater(this::registerJsBridge), 250L * multi, TimeUnit.MILLISECONDS);
    }

    protected Object jFedBridge = null;
    protected void registerJsBridge() {
        Future snapshotOfFuture = nextRegisterJsBridge;
        if (snapshotOfFuture != null) {
            //try to stop any next run of registerJsBridge (because we reschedule it later anyway). No problem if it doesn't stop.
            //any race conditions here are also not important.
            snapshotOfFuture.cancel(false);
        }

        //document.readyState === "complete" || document.readyState === "interactive"
        JSObject document = (JSObject) webEngine.executeScript("document");
        if (document == null) {
            registerJsBridgeLater();
            return;
        }
        String readyState = (String) document.getMember("readyState");
//        LOG.debug("registerJsBridge document.readyState = \""+readyState+"\"");
        if (readyState == null) {
            registerJsBridgeLater();
            return;
        }
        if (!readyState.equals("complete") && !readyState.equals("interactive")) {
            registerJsBridgeLater();
            return;
        }

        if (jFedBridge == null)
            jFedBridge = makeJavascriptBridge();

        JSObject win = (JSObject) webEngine.executeScript("window");
        Object existing = win.getMember("jfed");
        //noinspection ObjectEquality   //yes, we want the exact same object, not an equal one.
        if (existing != jFedBridge) {
            LOG.debug("registerJsBridge() -> Registering jFed in javascript");
            win.setMember("jfed", jFedBridge);

            //try again a bit later to be sure. Sometimes, it doesn't work the first time. Sometimes, it gets removed afterwards.
            jsBridgeBackoff = 1;
            registerJsBridgeLater();
        } else {
//            LOG.debug("registerJsBridge() -> jFed already registered in javascript");
            jsBridgeBackoff++;
            registerJsBridgeLater();
        }
    }

    protected Object makeJavascriptBridge() {
        return new JavaScriptBridge(this);
    }

    public HttpClientBuilder getHttpClientBuilder(URL url) {
        LOG.debug("getHttpClient url="+url);

        if (!geniUserProvider.isUserLoggedIn())
            return null;

        GeniUser user = geniUserProvider.getLoggedInGeniUser();
        BasicConnectionBuilderFactory connectionBuilderFactory = new BasicConnectionBuilderFactory();
        ConnectionBuilder connectionBuilder = connectionBuilderFactory.createConnectionBuilder();
        final JFedTrustStore conTrustStore = new JFedTrustStore(trustStore);
        Server server = null;
        //noinspection ConstantConditions
        if (server != null && server.hasCertificateChain()) {
            conTrustStore.addTrustedPemCertificateIfNotAdded(server.getCertificateChain());
            if (server.hasAllowedCertificateAlias()) {
                conTrustStore.addAllowedServerCertificateHostnameAlias(server.getAllowedCertificateAlias());
                LOG.debug("Using SSL conTrustStore with cert and with alias \""+server.getAllowedCertificateAlias()+"\" " +
                        "for server " + server.getName() + " (" + server.getId()+")");
            } else {
                LOG.debug("Using SSL conTrustStore with cert but without alias for server "+server.getName()+" ("+server.getId()+")");
            }
        }
//        else {
//            LOG.debug("Using SSL conTrustStore without cert or alias info for server "+server.getName()+" ("+server.getId()+")");
//        }
        connectionBuilder.useHttps(conTrustStore, null);
        connectionBuilder.useSslClientAuthentication(user.getClientCertificateChain(), user.getPrivateKey());
        connectionBuilder.setDebugInfo(new JFedConnection.DebugInfo(null, null));
        connectionBuilder.setUrl(url);
        HttpConnection con;
        try {
            con = connectionBuilder.buildHttpConnection();
        } catch (JFedException e) {
            LOG.error("Error building connection", e);
            return null;
        }

        HttpClientBuilder httpClientBuilder = ((BasicHttpConnection)con).getClientBuilder();
        return httpClientBuilder;
    }


    public class MyHttpUrlStreamHandler extends URLStreamHandler{
        @Override
        protected HttpURLConnection openConnection(URL u){
            LOG.debug("openConnection("+u.toExternalForm()+")");
            MyURLConnection q = new MyURLConnection(u);
            return q;
        }
    }

    private static class DebugOutputStream extends OutputStream {
        private final OutputStream delegate;

        private DebugOutputStream(OutputStream delegate) {
            this.delegate = delegate;
        }

        @Override
        public void write(int b) throws IOException {
            LOG.debug("write(b="+b+")");
            delegate.write(b);
        }

        @Override
        public void write(@NotNull byte[] b) throws IOException {
            LOG.debug("write(b[].length="+b.length+")");
            delegate.write(b);
        }

        @Override
        public void write(@NotNull byte[] b, int off, int len) throws IOException {
            LOG.debug("write(b[].length="+b.length+", off="+off+", len="+len+")");
            delegate.write(b, off, len);
        }

        @Override
        public void flush() throws IOException {
            LOG.debug("flush()");
            delegate.flush();
        }

        @Override
        public void close() throws IOException {
            LOG.debug("close()");
            delegate.close();
        }
    }

    private final Executor myHttpCLientExecutor = Executors.newSingleThreadExecutor();

    public class MyURLConnection extends HttpURLConnection
    {
        private final HttpClientBuilder httpClientBuilder;
        private HttpClient httpClient;
        private HttpResponse response;

        private final PipedOutputStream requestOutputStream;
        private final PipedInputStream requestInternalPipedInputStream;

        private final URI uri;

        private boolean responseError = false;

        protected MyURLConnection(URL url) {
            super(url);
            LOG.debug("MyURLConnection("+url+")");
            try {
                this.uri = url.toURI();
            } catch (URISyntaxException e) {
                throw new RuntimeException("Error creating URI from URL for \""+url.toExternalForm()+"\"", e);
            }
            this.httpClientBuilder = getHttpClientBuilder(url);
            this.requestOutputStream = new PipedOutputStream();
            this.requestInternalPipedInputStream = new PipedInputStream();
        }

        @Override
        public void disconnect() {
            LOG.debug("disconnect()");
            try {
                requestInternalPipedInputStream.close();
            } catch (IOException e) {
                //ignore any close exception
            }
            try {
                requestOutputStream.close();
            } catch (IOException e) {
                //ignore any close exception
            }
//            httpClient.?
//            response.?
        }

        @Override
        public boolean usingProxy() {
            LOG.debug("usingProxy()");
            return false;
        }

        @Override
        public void connect() throws IOException
        {
            LOG.debug("connect()");
            if (connected)
            {
                LOG.debug("already connected");
                return;
            }

            httpClientBuilder.setDefaultRequestConfig(RequestConfig.custom()
                    .setConnectTimeout(getConnectTimeout()) //setConnectTimeout = connection timeout in ms (default: 0 = infinite)
//                    .setConnectionRequestTimeout(getConnectTimeout()) //?
                    .setSocketTimeout(getReadTimeout()) //setSocketTimeout = read timeout in ms (default: 0 = infinite)
                    .build());

            this.httpClient = httpClientBuilder.build();


            HttpUriRequest httpCall;
            switch (method.toUpperCase()) {
                case "GET": { httpCall = new HttpGet(uri); break; }
                case "POST": { httpCall = new HttpPost(uri); break; }
                case "PUT": { httpCall = new HttpPut(uri); break; }
                case "DELETE": { httpCall = new HttpDelete(uri); break; }
                case "HEAD": { httpCall = new HttpHead(uri); break; }
                case "OPTIONS": { httpCall = new HttpOptions(uri); break; }
                case "TRACE": { httpCall = new HttpTrace(uri); break; }
                default: throw new ProtocolException("Unsupported HTTP method: " + method);
            }

            boolean onThread = false;
            if (httpCall instanceof HttpEntityEnclosingRequestBase) { //POST or PUT or ...
                BasicHttpEntity requestEntity = new BasicHttpEntity();
                if (fixedContentLength != -1)
                    requestEntity.setContentLength(fixedContentLength);
                if (fixedContentLengthLong != -1)
                    requestEntity.setContentLength(fixedContentLengthLong);
                requestEntity.setContent(requestInternalPipedInputStream);
                requestEntity.setChunked(chunkLength != -1); //we ignore the actual chunkLength, doesn't seem to be supported by httpClient
//                requestEntity.setContentType(???);
//                requestEntity.setContentEncoding(???);
                this.requestOutputStream.connect(requestInternalPipedInputStream);

                ((HttpEntityEnclosingRequestBase)httpCall).setEntity(requestEntity);

                onThread = true;
                LOG.debug("finished setup of requestEntity");
            }

            if (ifModifiedSince > 0) {
                Instant sinceDate = Instant.ofEpochMilli(ifModifiedSince);
                //If-Modified-Since expects RFC 822 (updated by RFC 1123)
                httpCall.setHeader("If-Modified-Since", java.time.format.DateTimeFormatter.RFC_1123_DATE_TIME.format(sinceDate));
            }

            Map<String, List<String>> requestHeaders = getRequestProperties();
            if (requestHeaders != null) {
                for (Map.Entry<String, List<String>> requestHeader : requestHeaders.entrySet()) {
                    httpCall.setHeader(requestHeader.getKey(), requestHeader.getValue().stream().collect(Collectors.joining(",")));
                }
            }

            connected = true;

            if (!onThread) {
                this.response = httpClient.execute(httpCall);
                LOG.debug("finished connect()");
            } else {
                Runnable exRun = () -> {
                    try {
                        this.response = httpClient.execute(httpCall);
                    } catch (IOException e) {
                        this.responseError = true;
                        throw new RuntimeException("httpClient.execute error", e);
                    }
                };
                myHttpCLientExecutor.execute(exRun);
                LOG.debug("finishing connect() in background thread");
            }
        }

        @Override
        public String getHeaderField(String name)
        {
            LOG.debug("getHeaderField("+name+")");
            waitForResponse();
            if (!response.containsHeader(name))
                return null;
            return response.getLastHeader(name).getValue();
        }

        @Override
        public String getContentType()
        {
            LOG.debug("getContentType()");
            waitForResponse();
            if (!response.containsHeader("Content-Type"))
                return null;
            return response.getLastHeader("Content-Type").getValue();
        }

        @Override
        public int getContentLength()
        {
            LOG.debug("getContentLength()");
            waitForResponse();
            return (int) response.getEntity().getContentLength();
        }

        @Override
        public long getContentLengthLong()
        {
            LOG.debug("getContentLengthLong()");
            waitForResponse();
            return response.getEntity().getContentLength();
        }

        @Override
        public boolean getDoInput()
        {
            LOG.debug("getDoInput()");
            return super.getDoInput();
        }

        @Override
        public InputStream getInputStream() throws IOException
        {
            connect();
            waitForResponse();
            LOG.debug("getInputStream()");
            return response.getEntity().getContent();
        }

        private void waitForResponse() {
            while (this.response == null && !responseError) {
                try {
                    LOG.debug("waitForResponse() is waiting");
                    Thread.sleep(100L);
                } catch (InterruptedException e) {
                    break; //break on interrupted
                }
            }
        }

        @Override
        public OutputStream getOutputStream() throws IOException
        {
            LOG.debug("getOutputStream()");
            return new DebugOutputStream(requestOutputStream);
        }

        @Override
        public java.security.Permission getPermission() throws IOException
        {
            LOG.debug("getPermission()");
            return super.getPermission();
        }

        @Override
        public String getHeaderFieldKey(int i) {
            LOG.debug("getHeaderFieldKey("+i+")");
            return super.getHeaderFieldKey(i);
        }

        @Override
        public void setFixedLengthStreamingMode(int i) {
            LOG.debug("setFixedLengthStreamingMode("+i+")");
            super.setFixedLengthStreamingMode(i);
        }

        @Override
        public void setFixedLengthStreamingMode(long l) {
            LOG.debug("setFixedLengthStreamingMode("+l+")");
            super.setFixedLengthStreamingMode(l);
        }

        @Override
        public void setChunkedStreamingMode(int i) {
            LOG.debug("setChunkedStreamingMode("+i+")");
            super.setChunkedStreamingMode(i);
        }

        @Override
        public String getHeaderField(int i) {
            LOG.debug("getHeaderField("+i+")");
            return super.getHeaderField(i);
        }

        @Override
        public void setInstanceFollowRedirects(boolean b) {
            LOG.debug("setInstanceFollowRedirects("+b+")");
            super.setInstanceFollowRedirects(b);
            if (!b)
                httpClientBuilder.disableRedirectHandling();
        }

        @Override
        public boolean getInstanceFollowRedirects() {
            LOG.debug("getInstanceFollowRedirects()");
            return super.getInstanceFollowRedirects();
        }

        @Override
        public void setRequestMethod(String s) throws ProtocolException {
            LOG.debug("setRequestMethod("+s+")");
            super.setRequestMethod(s);
        }

        @Override
        public String getRequestMethod() {
            LOG.debug("getRequestMethod()");
            return super.getRequestMethod();
        }

        @Override
        public int getResponseCode() throws IOException {
            LOG.debug("getResponseCode()");
            waitForResponse();
            return response.getStatusLine().getStatusCode();
//            return super.getResponseCode();
        }

        @Override
        public String getResponseMessage() throws IOException {
            LOG.debug("getResponseMessage()");
            waitForResponse();
            return response.getStatusLine().getReasonPhrase();
        }

        @Override
        public long getHeaderFieldDate(String s, long l) {
            LOG.debug("getHeaderFieldDate("+s+", "+l+")");
            //super calls getHeaderField so this will just work
            return super.getHeaderFieldDate(s, l);
        }

        @Override
        public InputStream getErrorStream() {
            LOG.debug("getErrorStream()");
            waitForResponse();
            return null; //TODO
            //probably something like this:
//            if (error && response.getEntity() != null)
//                return response.getEntity().getContent();
//            else
//                return null;
        }

        @Override
        public void setConnectTimeout(int i) {
            LOG.debug("setConnectTimeout("+i+")");
            super.setConnectTimeout(i);
        }

        @Override
        public int getConnectTimeout() {
            LOG.debug("getConnectTimeout() -> "+super.getConnectTimeout());
            return super.getConnectTimeout();
        }

        @Override
        public void setReadTimeout(int i) {
            LOG.debug("setReadTimeout("+i+")");
            super.setReadTimeout(i);
        }

        @Override
        public int getReadTimeout() {
            LOG.debug("getReadTimeout() -> "+super.getReadTimeout());
            return super.getReadTimeout();
        }

        @Override
        public URL getURL() {
            LOG.debug("getURL()");
            return super.getURL();
        }

        @Override
        public String getContentEncoding() {
            LOG.debug("getContentEncoding()");
            //just calls getHeaderField("content-encoding");
            return super.getContentEncoding();
        }

        @Override
        public long getExpiration() {
            LOG.debug("getExpiration()");
            //just calls getHeaderFieldDate("expires", 0);
            return super.getExpiration();
        }

        @Override
        public long getDate() {
            LOG.debug("getDate()");
            //just calls getHeaderFieldDate("date", 0);
            return super.getDate();
        }

        @Override
        public long getLastModified() {
            LOG.debug("getLastModified()");
            //just calls getHeaderFieldDate("last-modified", 0);
            return super.getLastModified();
        }

        @Override
        public Map<String, List<String>> getHeaderFields() {
            LOG.debug("getHeaderFields()");
            waitForResponse();
            Map<String, List<String>> res = new HashMap<>();
            for (Header header: response.getAllHeaders()) {
                res.put(header.getName(), Arrays.asList(header.getValue().split(",")));
            }
            return Collections.unmodifiableMap(res);
        }

        @Override
        public int getHeaderFieldInt(String s, int i) {
            LOG.debug("getHeaderFieldInt("+s+", "+i+")");
            //super uses getHeaderField(name); so ok to call
            return super.getHeaderFieldInt(s, i);
        }

        @Override
        public long getHeaderFieldLong(String s, long l) {
            LOG.debug("getHeaderFieldLong("+s+", "+l+")");
            //super uses getHeaderField(name); so ok to call
            return super.getHeaderFieldLong(s, l);
        }

        @Override
        public Object getContent() throws IOException {
            LOG.debug("getContent()");
            //super call should be ok to call as it uses getInputStream...
            return super.getContent();
        }

        @Override
        public Object getContent(Class[] classes) throws IOException {
            LOG.debug("getContent("+classes+")");
            //super call should be ok to call as it uses getInputStream...
            return super.getContent(classes);
        }

        @Override
        public void setDoInput(boolean b) {
            LOG.debug("setDoInput("+b+")");
            super.setDoInput(b);
        }

        @Override
        public void setDoOutput(boolean b) {
            LOG.debug("setDoOutput("+b+")");
            super.setDoOutput(b);
        }

        @Override
        public boolean getDoOutput() {
            LOG.debug("getDoOutput()");
            return super.getDoOutput();
        }

        @Override
        public void setAllowUserInteraction(boolean b) {
            LOG.debug("setAllowUserInteraction("+b+")");
            super.setAllowUserInteraction(b);
        }

        @Override
        public boolean getAllowUserInteraction() {
            LOG.debug("getAllowUserInteraction()");
            return super.getAllowUserInteraction();
        }

        @Override
        public void setUseCaches(boolean b) {
//            LOG.debug("setUseCaches("+b+")");
            super.setUseCaches(b);
        }

        @Override
        public boolean getUseCaches() {
//            LOG.debug("getUseCaches()");
            return super.getUseCaches();
        }

        @Override
        public void setIfModifiedSince(long l) {
            LOG.debug("setIfModifiedSince("+l+")");
            super.setIfModifiedSince(l);
        }

        @Override
        public long getIfModifiedSince() {
            LOG.debug("getIfModifiedSince()");
            return super.getIfModifiedSince();
        }

        @Override
        public boolean getDefaultUseCaches() {
            LOG.debug("getDefaultUseCaches()");
            return super.getDefaultUseCaches();
        }

        @Override
        public void setDefaultUseCaches(boolean b) {
            LOG.debug("setDefaultUseCaches("+b+")");
            super.setDefaultUseCaches(b);
        }

        @Override
        public void setRequestProperty(String s, String s1) {
            LOG.debug("setRequestProperty("+s+", "+s1+")");
            super.setRequestProperty(s, s1);
        }

        @Override
        public void addRequestProperty(String s, String s1) {
            LOG.debug("addRequestProperty("+s+", "+s1+")");
            super.addRequestProperty(s, s1);
        }

        @Override
        public String getRequestProperty(String s) {
            LOG.debug("getRequestProperty()");
            return super.getRequestProperty(s);
        }

        @Override
        public Map<String, List<String>> getRequestProperties() {
            LOG.debug("getRequestProperties() -> "+super.getRequestProperties());
            return super.getRequestProperties();
        }
    }
}
